package com.bluefairygames.java.libgdx.gameframework.object;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.badlogic.gdx.graphics.Camera;
import com.badlogic.gdx.math.Vector2;
import com.bluefairygames.java.libgdx.gameframework.enums.ObjectType;
import com.bluefairygames.java.libgdx.gameframework.object.component.Component;

/**
 * Base class for all objects in the game. A game object has an ID, name, and type, along with a
 * list of {@link Component components} and other attributes that define its physical location,
 * orientation, and other such properties.
 * 
 * @author Matt Parish
 */
public class GameObject {

	private int mId;
	private String mName;
	private ObjectType mType;

	protected Vector2 mPosition;
	protected float mRotation;

	protected Vector2 mSize;

	private List<Component> mComponents;

	private Map<String, Object> mProperties;

	public GameObject(int id, String name, ObjectType type) {

		mId = id;
		mName = name;
		mType = type;

		mPosition = new Vector2();
		mRotation = 0;

		mSize = new Vector2();

		mComponents = new ArrayList<Component>();
		mProperties = new HashMap<String, Object>();
	}

	public int getId() {

		return mId;
	}

	public String getName() {

		return mName;
	}

	public ObjectType getType() {

		return mType;
	}

	/*
	 * TODO: update the order of components after we add one so that things that need to update or
	 * render first are higher in the list of components
	 */
	public void addComponent(Component c) {

		mComponents.add(c);
	}

	public Vector2 getPosition() {

		return mPosition;
	}

	public void setPosition(float x, float y) {

		mPosition.set(x, y);
	}
	
	public void setPosition(Vector2 position) {
		
		mPosition = position;
	}

	public float getRotation() {

		return mRotation;
	}

	public void setRotation(float rotation) {

		mRotation = rotation;
	}

	public Vector2 getSize() {

		return mSize;
	}

	public void setSize(float width, float height) {

		mSize.set(width, height);
	}
	
	public void setSize(Vector2 size) {
		
		mSize = size;
	}

	/**
	 * @param name The name of the property being set.
	 * @param data The data object associated with this property.
	 */
	public void setProperty(String name, Object data) {

		mProperties.put(name, data);
	}

	/**
	 * @param name The name of the property to retrieve.
	 * @return The data object associated with the given property.
	 */
	public Object getProperty(String name) {

		return mProperties.get(name);
	}

	/**
	 * Updates all of the components associated with this object.
	 * 
	 * @param elapsedTime The time elapsed since the previous update.
	 */
	public void update(float elapsedTime) {

		for (Component c : mComponents) {
			c.update(elapsedTime);
		}
	}

	/**
	 * Render all of the components associated with this object that have visual representations.
	 * @param camera The camera to which the visual representations should be rendered.
	 */
	public void render(Camera camera) {

		for (Component component : mComponents) {
			
			component.render(camera);
		}
	}
}
