package com.bluefairygames.java.libgdx.gameframework.system;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import com.badlogic.gdx.Gdx;
import com.bluefairygames.java.libgdx.gameframework.events.Event;
import com.bluefairygames.java.libgdx.gameframework.events.EventListener;
import com.bluefairygames.java.libgdx.gameframework.object.GameObject;

public class EventManager {
	private static final String TAG = EventManager.class.getSimpleName();

	/**
	 * Static instance ensuring that the manager is only created once.
	 */
	private static EventManager sInstance = null;

	/**
	 * List of events awaiting delivery.
	 */
	private Queue<Event> mPendingEvents;

	/**
	 * Mapping used to deliver events to {@link GameObject}s explicitly designated as recipients by
	 * the sender.
	 */
	private Map<Class<? extends Event>, Map<String, List<EventListener>>> mRecipientMap;

	/**
	 * Mapping used for recipients who have registered for all events of a particular type.
	 */
	private Map<Class<? extends Event>, List<EventListener>> mInterestedPartyMap;

	public EventManager() {

		mPendingEvents = new LinkedList<Event>();
		mRecipientMap = new HashMap<Class<? extends Event>, Map<String, List<EventListener>>>();
		mInterestedPartyMap = new HashMap<Class<? extends Event>, List<EventListener>>();
	}

	public static EventManager get() {

		if (sInstance == null) {

			Gdx.app.log(TAG, "Creating static instance...");
			
			sInstance = new EventManager();
		}

		return sInstance;
	}

	/**
	 * <b>public void register(GameObject object, Class event, EventListener listener)</b><br>
	 * Register a particular {@link EventListener} within the given {@link GameObject} for a
	 * specific type of {@link Event}.
	 * 
	 * @param object The {@link GameObject} containing the listener to invoke when the event is
	 *        received.
	 * @param event The type of {@link Event} for which to listen.
	 * @param listener The {@link EventListener} that will be invoked when the event is received.
	 */
	public void register(GameObject object, Class<? extends Event> event, EventListener listener) {

		String objectName = object.getName();

		Map<String, List<EventListener>> listenerMap = mRecipientMap.get(event);
		if (listenerMap == null) {

			listenerMap = new HashMap<String, List<EventListener>>();
			mRecipientMap.put(event, listenerMap);
		}

		List<EventListener> listeners = listenerMap.get(objectName);
		if (listeners == null) {

			listeners = new ArrayList<EventListener>();
			listenerMap.put(objectName, listeners);

			Gdx.app.log(TAG, "Registering: " + objectName + " for " + event.getSimpleName());
		} else {

			Gdx.app.log(TAG, "Registering additional listener for " + event.getSimpleName()
				+ " in " + objectName);
		}

		listeners.add(listener);
	}

	/**
	 * <b>public void register(EventListener listener, Class event)</b><br>
	 * Register the given {@link EventListener} for all events of the given type.
	 * 
	 * @param listener The {@link EventListener} to register.
	 * @param event The {@link Event} for which to listen.
	 */
	public void register(EventListener listener, Class<? extends Event> event) {
		
		Gdx.app.log(TAG, "Registering " + listener.getClass().getSimpleName() + " for "
			+ event.getSimpleName() + " events.");

		List<EventListener> listeners = mInterestedPartyMap.get(event);
		if (listeners == null) {

			listeners = new ArrayList<EventListener>();
			mInterestedPartyMap.put(event, listeners);
		}

		listeners.add(0, listener);
	}

	/**
	 * <b>public void unregister(GameObject object, Class event, EventListener listener)</b><br>
	 * Unregister the given {@link EventListener} within the specified {@link GameObject} from the
	 * specified type of {@link Event}.
	 * 
	 * @param object The {@link GameObject} that the listener belongs to.
	 * @param event The type of {@link Event} from which to unregister.
	 * @param listener The {@link EventListener} currently registered for this type of event.
	 */
	public void unregister(GameObject object, Class<? extends Event> event, EventListener listener) {

		String objectName = object.getName();

		Map<String, List<EventListener>> listenerMap = mRecipientMap.get(event);
		if (listenerMap == null) {

			Gdx.app.error(TAG,
				objectName + " has never been registered for " + event.getSimpleName());
			return;
		}

		List<EventListener> listeners = listenerMap.get(objectName);
		if (listeners == null) {

			Gdx.app.log(TAG,
				objectName + " has no listeners currently registered for " + event.getSimpleName());
		} else {

			listeners.remove(listener);
		}
	}

	/**
	 * <b>public void unregister(EventListener listener, Class event)</b><br>
	 * Unregister the given {@link EventListener} from the specified {@link Event}.
	 * <p>
	 * This method can only be used to unregister listeners that had previously registered for
	 * <b>all</b> events of this type. Listeners registered from within a particular
	 * {@link GameObject} must use {@link #unregister(GameObject, Class, EventListener) this}
	 * version.
	 * 
	 * @param listener The {@link EventListener} to unregister.
	 * @param event The type of {@link Event} from which to unregister.
	 */
	public void unregister(EventListener listener, Class<? extends Event> event) {

		List<EventListener> listeners = mInterestedPartyMap.get(event);
		if (listeners == null) {

			Gdx.app.log(TAG, "No listeners currently registered for " + event.getSimpleName());
		} else {

			listeners.remove(listener);
		}
	}

	/**
	 * <b>public void unregister(GameObject object, Class event)</b><br>
	 * Unregister all listeners within the given {@link GameObject} from the specified {@link Event}
	 * .
	 * <p>
	 * Example use case is when a {@link GameObject} is leaving the game and will not return. (A
	 * player or character that has died).
	 */
	public void unregister(GameObject object, Class<? extends Event> event) {

		String objectName = object.getName();

		Map<String, List<EventListener>> listenerMap = mRecipientMap.get(event);
		if (listenerMap == null) {

			Gdx.app.error(TAG,
				objectName + " has never been registered for " + event.getSimpleName());
			return;
		}

		List<EventListener> listeners = listenerMap.get(objectName);
		if (listeners == null) {

			Gdx.app.log(TAG,
				objectName + " has no listeners currently registered for " + event.getSimpleName());
		} else {

			for (EventListener listener : listeners) {

				listeners.remove(listener);
			}

			listenerMap.remove(objectName);
		}
	}

	public void dispatch(Event event) {
		
		Gdx.app.log(TAG, "Dispatching " + event.getClass().getSimpleName() + " event.");

		mPendingEvents.add(event);
	}

	public void dispatchImmediate(Event event) {
		
		String eventType = event.getClass().getSimpleName();

		// First dispatch the event to interested parties registered for all events of this type
		List<EventListener> interestedParties = mInterestedPartyMap.get(event.getClass());
		if (interestedParties != null) {

			Gdx.app.log(TAG, "Dispatching " + eventType + " event to interested parties...");

			for (EventListener listener : interestedParties) {

				listener.onEvent(event);
			}
		}

		// Now dispatch to all of the recipient GameObjects explicitly designated by the sender
		List<GameObject> recipients = event.getRecipients();
		Map<String, List<EventListener>> listenerMap = mRecipientMap.get(event.getClass());
		if (listenerMap != null) {

			Gdx.app.log(TAG, "Dispatching " + eventType + " event to designated recipients...");

			for (GameObject recipient : recipients) {

				List<EventListener> listeners = listenerMap.get(recipient.getName());
				if (listeners != null) {

					for (EventListener listener : listeners) {
						listener.onEvent(event);
					}
				}
			}
		}
	}

	/**
	 * <b>public void update(float elapsedTime)</b><br>
	 * Dispatch all events currently awaiting delivery.
	 * 
	 * @param elapsedTime The elapsed time since the previous update.
	 */
	public void update(float elapsedTime) {

		if(!mPendingEvents.isEmpty()) {
			
			// Make a copy of the queue and clear the member queue so that the manager can
			// catch any new events that are generated while updating.
			List<Event> pendingEvents = new LinkedList<Event>(mPendingEvents);
			mPendingEvents.clear();
	
			Gdx.app.log(TAG, "Dispatching events...");
	
			List<Event> delayedEvents = new LinkedList<Event>();
			for (Event event : pendingEvents) {
	
				// If the event is for a future time, add it to the delayed events list.
				if (event.getDispatchTime() > elapsedTime) {
	
					Gdx.app.log(TAG, event.getClass().getSimpleName() + 
						" event is for a future time: retaining event.");
	
					delayedEvents.add(event);
				} else {
	
					dispatchImmediate(event);
				}
			}
	
			mPendingEvents.addAll(delayedEvents);
		}
	}

	public void finish() {

		sInstance = null;
	}
}
